import os
import base64
import json

import numpy as np
import faiss
from google.cloud import storage
from google.cloud import bigtable
from google.cloud.bigtable import row_filters

from .query_result import IdsQR

def download_gcs_dir(project_id=None, bucket_name=None, prefix=None, dl_dir=None):
    storage_client = storage.Client(project_id)
    bucket = storage_client.get_bucket(bucket_name)
    blobs = bucket.list_blobs(prefix=prefix)  # Get list of files
    for blob in blobs:
        filename = blob.name.replace(prefix, '') 
        print('downloading %s' % (filename))
        blob.download_to_filename(dl_dir + filename)

def parseMetadata(md):
    try:
        return json.loads(md.decode("utf-8"))
    except:
        return md.decode("utf-8")

stopwords_sets = {
    'english': {"it", "its", "itself", "what", "which", "who", "whom", "this", "that", "these", "those", "am", "is", "are", "was", "were", "be", "been", "being", "have", "has", "had", "having", "do", "does", "did", "doing", "a", "an", "the", "and", "but", "if", "or", "because", "as", "of", "at", "by", "for", "with", "to", "from", "in", "then", "all", "any", "both", "each", "few", "more", "most", "other", "some", "such", "no", "nor", "not", "only", "own", "same", "so", "than", "too", "very", "s", "t", "can", "will", "just", "don", "should", "now"}
}



class Index(object):
    """
    The data. Contains inverted index postings, faiss indices.
    """
    def __init__(self, name, 
        project_id='',
        bt_instance_id='',
        bt_table_id='',
        stopwords='english', 
        stopword_threshold=None):
        self.name = name
        self.project_id = project_id
        self.bt_instance_id = bt_instance_id
        self.bt_table_id = bt_table_id
        self.stopword_threshold = stopword_threshold
        self.stopwords = stopwords_sets[stopwords] if type(stopwords) is str else stopwords
        
        self.bt_client = bigtable.Client(project=self.project_id)
        self.bt_instance = self.bt_client.instance(self.bt_instance_id)
        self.bt_table = self.bt_instance.table(self.bt_table_id)
        self.bt_row_filter = row_filters.CellsColumnLimitFilter(1)

        self.index_dir = 'data/%s' % (self.name)
        self.ii = {}
        self.postings_dir = 'data/%s/postings/' % (self.name)
        self.faiss_index = None
    
    def downloadPostings(self, postings_uri):
        
        if not os.path.exists('data'):
            os.mkdir('data')
        if not os.path.exists(self.index_dir):
            os.mkdir(self.index_dir)
        
        
        if not os.path.exists(self.postings_dir):
            postings_uri_prefix = postings_uri.replace('gs://', '')
            bucket_name = postings_uri_prefix.split('/')[0]
            path = '/'.join(postings_uri_prefix.split('/')[1:])
            # download postings files
            os.mkdir(self.postings_dir)
            download_gcs_dir(
                project_id=self.project_id,
                bucket_name=bucket_name,
                prefix=path,
                dl_dir=self.postings_dir
            )
    
    def loadPostingsFromFile(self):
        
        for fn in os.listdir(self.postings_dir):
            with open(os.path.join(self.postings_dir, fn), 'r') as f:
                for line in f:
                    p = line.split('|||')
                    row_key = p[0]
                    if len(p) > 1:
                        a = None
                        if row_key.startswith('score.'):
                            # let's have some logic in here to figure out which QR is best
                            # some scores might be sparse (searchwise_scores) for instance
                            # while some may have a non-zero value for every product
                            # TODO: revisit
                            a = np.frombuffer(base64.b64decode(p[1]), dtype=np.float16)
                        else:
                            # if we maintain a sorted postings list, we can optimize filtering greatly
                            # (n^2 ->  n)
                            a = np.sort(np.frombuffer(base64.b64decode(p[1]), dtype=np.int32))
                        if len(row_key.split('.')) < 2:
                            print('row_key not correct format: %s' % (row_key))
                        else:
                            field = row_key.split('.')[0]
                            value = '.'.join(row_key.split('.')[1:])
                            if value not in self.stopwords:
                                self.putPostings(field, value, a)
                    else:
                        print(line)
    
    def load(self, faiss_filename, postings_uri):
        self.ii = {}
        self.downloadPostings(postings_uri)
        self.loadPostingsFromFile()
    
    def btLookup(self, row_key):
        row = self.bt_table.read_row(row_key.encode('utf-8'), self.bt_row_filter)
        column_family_id = 'p'
        column_id = 'p'.encode('utf-8')
        value = row.cells[column_family_id][column_id][0].value
        return value
    
    def getProductId(self, i):
        row_key = 'i.%s' % (i)
        response = self.btLookup(row_key).decode("utf-8")
        return response

    def getMetadata(self, i):
        row_key = 'p.%s' % (i)
        response = parseMetadata(self.btLookup(row_key))
        return response

    def getRemotePostings(self, field, value):
        row_key = '%s.%s' % (field, base64.b64encode(value.encode()).decode())
        value = self.btLookup(row_key)
        return IdsQR().fromIds(np.array(np.frombuffer(value, dtype=np.int32)))
    
    def getPostings(self, field, value):
        key = '%s.%s' % (field, base64.b64encode(value.encode()).decode())
        if key in self.ii:
            return self.ii[key]
        else:
            return self.getRemotePostings(field, value)
    
    def putPostings(self, field, value, ids):
        key = '%s.%s' % (field, value)
        self.ii[key] = IdsQR().fromIds(ids)



class QueryContext(object):
    def __init__(self, index):
        self.index = index
        self.setting = None
        self.filter_ids = None
        self.not_filter_ids = None
    
    def setSetting(self, setting):
        self.setting = setting




from .query_context import QueryContext
from .query_components import Match, Terms, Term, And, Or
from .query_result import IdsQR, IdsScoresQR
from .combine_scores.wrapped import sorted_in1d, combine_must_should

import numpy as np
import time

def filterHandler(value, qc):
    assert type(value) is list
    assert False not in [type(obj) is dict and len(obj) == 1 for obj in value] # must contain only queries
    qc.setSetting('filter')
    clauses = [query_part_handlers[list(obj.keys())[0]](obj[list(obj.keys())[0]], qc) for obj in value]
    return And(qc)._filter(clauses)

def mustNotHandler(value, qc):
    assert type(value) is list
    assert False not in [type(obj) is dict and len(obj) == 1 for obj in value] # must contain only queries
    qc.setSetting('filter')
    clauses = [query_part_handlers[list(obj.keys())[0]](obj[list(obj.keys())[0]], qc) for obj in value]
    return Or(qc)._filter(clauses)

def mustHandler(value, qc):
    assert type(value) is list
    assert False not in [type(obj) is dict and len(obj) == 1 for obj in value] # must contain only queries
    qc.setSetting('score')
    clauses = [query_part_handlers[list(obj.keys())[0]](obj[list(obj.keys())[0]], qc) for obj in value]
    return And(qc)._score(clauses)

def shouldHandler(value, qc):
    assert type(value) is list
    assert False not in [type(obj) is dict and len(obj) == 1 for obj in value] # must contain only queries
    qc.setSetting('score')
    clauses = [query_part_handlers[list(obj.keys())[0]](obj[list(obj.keys())[0]], qc) for obj in value]
    return Or(qc)._score(clauses)


def boolHandler(value, qc):

    # _filter = query_part_handlers['filter'](value.get('filter'), qc) if value.get('filter') else None
    # _must_not = query_part_handlers['must_not'](value.get('must_not'), qc) if value.get('must_not') else None
    # _must = query_part_handlers['must'](value.get('must'), qc) if value.get('must') else None
    # _should = query_part_handlers['should'](value.get('should'), qc) if value.get('should') else None

    has_filter = value.get('filter') is not None
    has_must_not = value.get('must_not') is not None
    has_must = value.get('must') is not None
    has_should = value.get('should') is not None

    has_any_filter = has_filter or has_must_not
    has_any_score = has_must or has_should

    has_filter_and_must_not = has_filter and has_must_not
    has_must_and_should = has_must and has_should

    if has_any_filter:

        combined_filter_ids = None
        combined_filter_invert = False
        if has_filter:
            if has_must_not:
                # if it has both, we can combine them to a single ids list
                _filter = query_part_handlers['filter'](value.get('filter'), qc)
                _must_not = query_part_handlers['must_not'](value.get('must_not'), qc)

                # using sorted_in1d with invert=True returns the ids in the first arg that are not in the second
                combined_filter_ids = _filter.getIds()[sorted_in1d(_filter.getIds(), _must_not.getIds(), invert=True, assume_sorted=True)]
                combined_filter_invert = False
            else:
                _filter = query_part_handlers['filter'](value.get('filter'), qc)
                combined_filter_ids = _filter.getIds()
                combined_filter_invert = False
        else:
            combined_filter_ids = query_part_handlers['must_not'](value.get('must_not'), qc).getIds()
            combined_filter_invert = True

        if has_must_and_should:
            _should = query_part_handlers['should'](value.get('should'), qc)
            _must = query_part_handlers['must'](value.get('must'), qc)

            filtered_must_mask = sorted_in1d(_must.getIds(), combined_filter_ids, invert=combined_filter_invert)
            _filtered_must_qr = IdsScoresQR(_must.getIds()[filtered_must_mask], _must.getScores()[filtered_must_mask])
            fm_s_ids, fm_s_scores = combine_must_should(_filtered_must_qr.getIds(), _filtered_must_qr.getScores(), _should.getIds(), _should.getScores(), assume_sorted=True)
            return IdsScoresQR(fm_s_ids, fm_s_scores)
        elif has_must:
            _must = query_part_handlers['must'](value.get('must'), qc)

            filtered_must_mask = sorted_in1d(_must.getIds(), combined_filter_ids, invert=combined_filter_invert, assume_sorted=True)
            _filtered_must_qr = IdsScoresQR(_must.getIds()[filtered_must_mask], _must.getScores()[filtered_must_mask])
            return _filtered_must_qr
        elif has_should:
            _should = query_part_handlers['should'](value.get('should'), qc)

            filtered_should_mask = sorted_in1d(_should.getIds(), combined_filter_ids, invert=combined_filter_invert, assume_sorted=True)
            _filtered_should_qr = IdsScoresQR(_should.getIds()[filtered_should_mask], _should.getScores()[filtered_should_mask])
            return _filtered_should_qr
        else:
            return IdsQR().fromIds(combined_filter_ids)
          
    else:
        # there are not filter / must_not
        # if there is must and should we combine, otherwise, return the one present
        if has_must_and_should:
            # combine must and should
            _should = query_part_handlers['should'](value.get('should'), qc)
            _must = query_part_handlers['must'](value.get('must'), qc)
            fm_s_ids, fm_s_scores = combine_must_should(_must.getIds(), _must.getScores(), _should.getIds(), _should.getScores(), assume_sorted=True)
            return IdsScoresQR(fm_s_ids, fm_s_scores)
        elif has_must:
            _must = query_part_handlers['must'](value.get('must'), qc)
            return _must
        else:
            _should = query_part_handlers['should'](value.get('should'), qc)
            return _should
    

def termHandler(value, qc):
    assert type(value) is dict and len(value) == 1
    
    field = [key for key in value][0]
    o = value[field]
    term = o.get('value')
    assert term is not None
    boost = o.get('boost')
    
    if qc.setting == 'filter':
        assert boost is None
        return Term(qc)._filter(field, term)
    elif qc.setting == 'score':
        return Term(qc)._score(field, term, score=boost)
    else:
        assert False

def matchHandler(value, qc):
    assert type(value) is dict and len(value) == 1
    field = [key for key in value][0]
    terms = value[field]
    
    if qc.setting == 'filter':
        return Match(qc)._filter(field, terms)
    elif qc.setting == 'score':
        return Match(qc)._score(field, terms)
    else:
        assert False
    

query_part_handlers = {
    'bool': boolHandler,
    'filter': filterHandler,
    'must': mustHandler,
    'should': shouldHandler,
    'must_not': mustNotHandler,
    'term': termHandler,
    'match': matchHandler
}

def doQuery(query,
        index=None,
        limit=10,
        offset=0,
        min_score=0,
        return_scores=True,
        return_doc_ids=True,
        return_product_ids=False,
        return_products=False):
    start_time = time.time()

    qc = QueryContext(index)
    qc.setSetting('score')
    assert len(query) == 1
    entrypoint = [key for key in query][0]
    ids, scores = query_part_handlers[entrypoint](query[entrypoint], qc).getTopN(limit + offset)

    query_done_time = time.time()

    start = min(offset, ids.size)
    end = min(offset + limit, ids.size)

    doc_ids = [int(_id) for _id in ids[start:end]]

    response = {}
    if return_doc_ids:
        response['docs'] = doc_ids
    if return_scores:
        response['scores'] = [float(score) for score in scores[start:end]]
    if return_product_ids:
        response['product_ids'] = [index.getProductId(i) for i in doc_ids]
    if return_products:
        response['products'] = [index.getMetadata(i) for i in doc_ids]
    
    response_done_time = time.time()

    response['query_took'] = query_done_time - start_time
    response['response_took'] = response_done_time - start_time

    return response


import numpy as np
# import heapq

# from .combine_scores import sum_scores

class QueryResult(object):
    def __init__(self):
        self.ids = None
        self.scores = None
    
    def getIds(self):
        return self.ids
    
    def getScores(self):
        return self.scores
    
    def getTopN(self, n, return_scores=True):
        n_clip = min(n, self.getIds().size)
        
        top_indices = np.flip(np.argpartition(self.getScores(), -n_clip)[-n_clip:])
        top_scores = np.take(self.getScores(), top_indices)
        top_scores_indices = np.flip(np.argsort(top_scores))
        top_scores_sorted = np.take(top_scores, top_scores_indices)
        top_ids_sorted = np.take(np.take(self.getIds(), top_indices), top_scores_indices)
        
        if return_scores:
            return (top_ids_sorted, top_scores_sorted)
        else:
            return (top_ids)

class IdsScoreQR(QueryResult):
    """
    A type of query result consisting of a numpy array of ids (integers)
    and a float16 'score' (array([1,2,4]), 1.0). 
    ex. This QR is returned by the Term query in score setting and is how postings are stored
    """
    def __init__(self):
        super().__init__()
        self.score = 1.0
    
    def getScores(self):
        if self.scores is not None:
            return self.scores
        else:
            self.scores = np.full(self.ids.size, self.score, dtype=np.float32)
            return self.scores
    
    # maybe we don't want this one
    def fromQR(self, qr):
        self.ids = qr.getIds() # might need to copy
        self.scores = qr.getScores()
        return self
    
    def fromQRScore(self, qr, score):
        self.ids = qr.getIds() # might need to copy
        self.score = score
        return self
    
    def getTopN(self, n, return_scores=True):
        top_ids = self.getIds()[0:n]
        if return_scores:
            top_scores = np.full(top_ids.size, self.score, dtype=np.float32)
            return (top_ids, top_scores)
        else:
            return (top_ids)

class IdsScoresQR(QueryResult):
    """
    A type of query result with np arrays for ids and scores
    """
    def __init__(self, ids, scores):
        self.ids = ids
        self.scores = scores
    
    def fromIdsScores(self, ids, scores):
        self.ids = ids
        self.scores = scores

class IdsQR(QueryResult):
    """
    A type of query result consisting of a numpy array of ids (integers)
    ex. This QR is returned by the Term query in filter setting
    """
    def __init__(self):
        super().__init__()
    
    def fromQR(self, qr):
        assert isinstance(qr, QueryResult)
        self.ids = qr.getIds() # might need to copy
        return self

    def fromIds(self, ids):
        assert type(ids) is np.ndarray
        self.ids = ids
        return self
    
    def getScores(self):
        if self.scores is None:
            self.scores = np.zeros(self.getIds().size, dtype=np.float32)
        return self.scores
    
    def getTopN(self, n, return_scores=True):
        top_ids = self.getIds()[0:n]
        if return_scores:
            top_scores = np.full(top_ids.size, 0, dtype=np.float32)
            return (top_ids, top_scores)
        else:
            return (top_ids)



from setuptools import setup, Extension
from Cython.Distutils import build_ext
import numpy as np

NAME = "mark-search"
VERSION = "0.0.18"
DESCR = "!"
REQUIRES = [
    'numpy', 
    'cython',
    'faiss',
    'google-cloud-bigtable',
    'google-cloud-storage',
    'gensim'
]

AUTHOR = "Mark Hudson"
EMAIL = "mark.cd.hudson@gmail.com"

LICENSE = "Apache 2.0"

SRC_DIR = "mark_search"
PACKAGES = [SRC_DIR, SRC_DIR + ".combine_scores"]

ext_1 = Extension(SRC_DIR + ".combine_scores.wrapped",
                  [SRC_DIR + "/combine_scores/wrapped.pyx"],
                  libraries=[],
                  include_dirs=[np.get_include()])


EXTENSIONS = [ext_1]

if __name__ == "__main__":
    setup(install_requires=REQUIRES,
          packages=PACKAGES,
          zip_safe=False,
          name=NAME,
          version=VERSION,
          description=DESCR,
          author=AUTHOR,
          author_email=EMAIL,
          license=LICENSE,
          cmdclass={"build_ext": build_ext},
          ext_modules=EXTENSIONS
          )



import functools
import math
import gensim

import numpy as np

from .query_result import IdsQR, IdsScoresQR, IdsScoreQR
from mark_search.combine_scores import sorted_multi_sum_scores_must, sorted_intersect1d, sorted_multi_sum_scores, sorted_multi_union1d


class QueryComponent(object):
    """
    Abstract base class for query components such as term, match, and, or...
    Args:
        name (str): component name.
        context (:obj:`QueryContext`): Contains information about how this component
            was called. Might include limit, offset, and score/filter mode
        candidates (:obj:`QueryResult`): Results of subqueries. Might be one of many forms.
    """
    def __init__(self, name, context, candidates):
        self.name = name
        self.context = context
        self.candidates = candidates
    
    # called in the filter context
    def _filter(self):
        pass
    
    # called in the score context
    def _score(self):
        pass

# in filter setting
# returns an IdsQR containing the intersection of the candidate list qr's
# in score setting
# adds up all scores 
class And(QueryComponent):
    def __init__(self, context):
        super().__init__('and', context, None)
    
    def _filter(self, candidate_list):
        candidate_list.sort(key=lambda x: x.getIds().size)
        # maybe put some logic in here to decide whether to do a linear
        # merge or a divide and conquer merge. depending on min_size:max_size:length
        return functools.reduce(lambda a, b: IdsQR().fromIds(sorted_intersect1d(a.getIds(), b.getIds(), assume_sorted=True)), candidate_list)

    def _score(self, candidate_list):
        ids_list = [e.getIds() for e in candidate_list]
        scores_list = [e.getScores() for e in candidate_list]
        ids, scores = sorted_multi_sum_scores_must(ids_list, scores_list, assume_sorted=True)
        return IdsScoresQR(ids, scores)


# in filter setting
# returns an IdsQR containing the union of the candidate list qr's
# in score setting
# adds up all scores 
class Or(QueryComponent):
    def __init__(self, context):
        super().__init__('or', context, None)

    def _filter(self, candidate_list):
        ids_list = [e.getIds() for e in candidate_list]
        return IdsQR().fromIds(sorted_multi_union1d(ids_list, assume_sorted=True))

    def _score(self, candidate_list):
        ids_list = [e.getIds() for e in candidate_list]
        scores_list = [e.getScores() for e in candidate_list]
        ids, scores = sorted_multi_sum_scores(ids_list, scores_list, assume_sorted=True)
        return IdsScoresQR(ids, scores)


class Term(QueryComponent):
    def __init__(self, context):
        super().__init__('term', context, None)
    
    def _filter(self, field, term):
        return self.context.index.getPostings(field, term)
    
    def _score(self, field, term, score=None):
        postings = self.context.index.getPostings(field, term)
        return IdsScoreQR().fromQRScore(postings, score or 1 / math.log(2 + postings.getIds().size))

class Terms(QueryComponent):
    def __init__(self, context):
        super().__init__('terms', context, None)
    
    def tokenize(self, s):
        return gensim.utils.simple_preprocess(s)
    
    def _filter(self, field, terms):
        tokens = []
        if type(terms) is str:
            tokens = self.tokenize(terms)
        elif type(terms) is list:
            tokens = terms
        term_queries = [Term(self.context)._filter(field, term) for term in tokens]
        or_query_result = Or(self.context)._filter(term_queries)
        return or_query_result
    
    def _score(self, field, terms):
        tokens = []
        if type(terms) is str:
            tokens = self.tokenize(terms)
        elif type(terms) is list:
            tokens = terms
        term_queries = [Term(self.context)._score(field, term) for term in tokens]
        print(term_queries)
        or_query_result = Or(self.context)._score(term_queries)
        return or_query_result

class Match(QueryComponent):
    def __init__(self, context):
        super().__init__('match', context, None)
    
    def tokenize(self, s):
        return gensim.utils.simple_preprocess(s)
    
    def _filter(self, field, terms):
        tokens = []
        if type(terms) is str:
            tokens = self.tokenize(terms)
        elif type(terms) is list:
            tokens = terms
        term_queries = [Term(self.context)._filter(field, term) for term in tokens]
        or_query_result = Or(self.context)._filter(term_queries)
        return or_query_result
    
    def _score(self, field, terms):
        tokens = []
        if type(terms) is str:
            tokens = self.tokenize(terms)
        elif type(terms) is list:
            tokens = terms
        term_queries = [Term(self.context)._score(field, term) for term in tokens if term not in self.context.index.stopwords]
        or_query_result = Or(self.context)._score([tq for tq in term_queries if tq.getIds().size < (self.context.index.stopword_threshold or 100000000)])
        return or_query_result

